@@ -1,6 +1,7 @@
 /*
  * Copyright (C) 2004-2012 Free Software Foundation, Inc.
  * Copyright (C) 2001,2002 Paul Sheer
+ * Copyright (C) 2016-2018 Red Hat, Inc.
  * Portions Copyright (C) 2002,2003 Nikos Mavrogiannopoulos
  *
  * This file is part of GnuTLS.
@@ -34,7 +35,6 @@
 #include <string.h>
 #include <gnutls/gnutls.h>
 #include <gnutls/dtls.h>
-#include <gnutls/openpgp.h>
 #include <sys/time.h>
 #include <sys/select.h>
 #include <fcntl.h>
@@ -61,28 +61,33 @@ static int debug = 0;
 unsigned int verbose = 1;
 static int nodb;
 static int noticket;
+static int earlydata;
 int require_cert;
 int disable_client_cert;
 
 const char *psk_passwd = NULL;
 const char *srp_passwd = NULL;
 const char *srp_passwd_conf = NULL;
-const char *pgp_keyring = NULL;
-const char *pgp_keyfile = NULL;
-const char *pgp_certfile = NULL;
-const char *x509_keyfile = NULL;
-const char *x509_certfile = NULL;
-const char *x509_dsakeyfile = NULL;
-const char *x509_dsacertfile = NULL;
-const char *x509_ecckeyfile = NULL;
-const char *x509_ecccertfile = NULL;
+const char **x509_keyfile = NULL;
+const char **x509_certfile = NULL;
+unsigned x509_certfile_size = 0;
+unsigned x509_keyfile_size = 0;
 const char *x509_cafile = NULL;
 const char *dh_params_file = NULL;
 const char *x509_crlfile = NULL;
 const char *priorities = NULL;
-const char *status_response_ocsp = NULL;
+
+const char **ocsp_responses = NULL;
+unsigned ocsp_responses_size = 0;
+
+const char *sni_hostname = NULL;
+int sni_hostname_fatal = 0;
+
+const char **alpn_protos = NULL;
+unsigned alpn_protos_size = 0;
 
 gnutls_datum_t session_ticket_key;
+gnutls_anti_replay_t anti_replay;
 static void tcp_server(const char *name, int port);
 
 /* end of globals */
@@ -113,7 +118,7 @@ gnutls_psk_server_credentials_t psk_cred = NULL;
 gnutls_anon_server_credentials_t dh_cred = NULL;
 gnutls_certificate_credentials_t cert_cred = NULL;
 
-const int ssl_session_cache = 128;
+const int ssl_session_cache = 2048;
 
 static void wrap_db_init(void);
 static void wrap_db_deinit(void);
@@ -121,6 +126,8 @@ static int wrap_db_store(void *dbf, gnutls_datum_t key,
 			 gnutls_datum_t data);
 static gnutls_datum_t wrap_db_fetch(void *dbf, gnutls_datum_t key);
 static int wrap_db_delete(void *dbf, gnutls_datum_t key);
+static int anti_replay_db_add(void *dbf, time_t exp, const gnutls_datum_t *key,
+			      const gnutls_datum_t *data);
 
 static void cmd_parser(int argc, char **argv);
 
@@ -135,7 +142,9 @@ LIST_TYPE_DECLARE(listener_item, char *http_request; char *http_response;
 		  int listen_socket; int fd;
 		  gnutls_session_t tls_session;
 		  int handshake_ok;
+		  int close_ok;
 		  time_t start;
+		  int earlydata_eof;
     );
 
 static const char *safe_strerror(int value)
@@ -152,7 +161,8 @@ static void listener_free(listener_item * j)
 	free(j->http_request);
 	free(j->http_response);
 	if (j->fd >= 0) {
-		gnutls_bye(j->tls_session, GNUTLS_SHUT_WR);
+		if (j->close_ok)
+			gnutls_bye(j->tls_session, GNUTLS_SHUT_WR);
 		shutdown(j->fd, 2);
 		close(j->fd);
 		gnutls_deinit(j->tls_session);
@@ -238,38 +248,6 @@ static void read_dh_params(void)
 
 }
 
-static char pkcs3[] =
-    "-----BEGIN DH PARAMETERS-----\n"
-    "MIGGAoGAtkxw2jlsVCsrfLqxrN+IrF/3W8vVFvDzYbLmxi2GQv9s/PQGWP1d9i22\n"
-    "P2DprfcJknWt7KhCI1SaYseOQIIIAYP78CfyIpGScW/vS8khrw0rlQiyeCvQgF3O\n"
-    "GeGOEywcw+oQT4SmFOD7H0smJe2CNyjYpexBXQ/A0mbTF9QKm1cCAQU=\n"
-    "-----END DH PARAMETERS-----\n";
-
-static int static_dh_params(void)
-{
-	gnutls_datum_t params = { (void *) pkcs3, sizeof(pkcs3) };
-	int ret;
-
-	if (gnutls_dh_params_init(&dh_params) < 0) {
-		fprintf(stderr, "Error in dh parameter initialization\n");
-		exit(1);
-	}
-
-	ret = gnutls_dh_params_import_pkcs3(dh_params, &params,
-					    GNUTLS_X509_FMT_PEM);
-
-	if (ret < 0) {
-		fprintf(stderr, "Error parsing dh params: %s\n",
-			safe_strerror(ret));
-		exit(1);
-	}
-
-	printf
-	    ("Set static Diffie-Hellman parameters, consider --dhparams.\n");
-
-	return 0;
-}
-
 static int
 get_params(gnutls_session_t session, gnutls_params_type_t type,
 	   gnutls_params_st * st)
@@ -300,36 +278,125 @@ int ret;
 		if (!require_cert && gnutls_certificate_get_peers(session, &size) == NULL)
 			return 0;
 
-		if ((require_cert || ENABLED_OPT(VERIFY_CLIENT_CERT)) && cert_verify(session, NULL, NULL) == 0) {
-			do {
-				ret = gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_ACCESS_DENIED);
-			} while(ret == GNUTLS_E_INTERRUPTED || ret == GNUTLS_E_AGAIN);
+		if (ENABLED_OPT(VERIFY_CLIENT_CERT)) {
+			if (cert_verify(session, NULL, NULL) == 0) {
+				do {
+					ret = gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_ACCESS_DENIED);
+				} while(ret == GNUTLS_E_INTERRUPTED || ret == GNUTLS_E_AGAIN);
 
-			j->http_state = HTTP_STATE_CLOSING;
-			return -1;
+				j->http_state = HTTP_STATE_CLOSING;
+				return -1;
+			}
+		} else {
+			printf("- Peer's certificate was NOT verified.\n");
 		}
 	}
 	return 0;
 }
 
+/* callback used to verify if the host name advertised in client hello matches
+ * the one configured in server
+ */
+static int
+post_client_hello(gnutls_session_t session)
+{
+	int ret;
+	/* DNS names (only type supported) may be at most 256 byte long */
+	char *name;
+	size_t len = 256;
+	unsigned int type;
+	int i;
+
+	name = malloc(len);
+	if (name == NULL)
+		return GNUTLS_E_MEMORY_ERROR;
+
+	for (i=0; ; ) {
+		ret = gnutls_server_name_get(session, name, &len, &type, i);
+		if (ret == GNUTLS_E_SHORT_MEMORY_BUFFER) {
+			char *new_name;
+			new_name = realloc(name, len);
+			if (new_name == NULL) {
+				ret = GNUTLS_E_MEMORY_ERROR;
+				goto end;
+			}
+			name = new_name;
+			continue; /* retry call with same index */
+		}
+
+		/* check if it is the last entry in list */
+		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE)
+			break;
+		i++;
+		if (ret != GNUTLS_E_SUCCESS)
+			goto end;
+		/* unknown types need to be ignored */
+		if (type != GNUTLS_NAME_DNS)
+			continue;
+
+		if (strlen(sni_hostname) != len)
+			continue;
+		/* API guarantees that the name of type DNS will be null terminated */
+		if (!strncmp(name, sni_hostname, len)) {
+			ret = GNUTLS_E_SUCCESS;
+			goto end;
+		}
+	};
+	/* when there is no extension, we can't send the extension specific alert */
+	if (i == 0) {
+		fprintf(stderr, "Warning: client did not include SNI extension, using default host\n");
+		ret = GNUTLS_E_SUCCESS;
+		goto end;
+	}
+
+	if (sni_hostname_fatal == 1) {
+		/* abort the connection, propagate error up the stack */
+		ret = GNUTLS_E_UNRECOGNIZED_NAME;
+		goto end;
+	}
+
+	fprintf(stderr, "Warning: client provided unrecognized host name\n");
+	/* since we just want to send an alert, not abort the connection, we
+	 * need to send it ourselves
+	 */
+	do {
+		ret = gnutls_alert_send(session,
+					GNUTLS_AL_WARNING,
+					GNUTLS_A_UNRECOGNIZED_NAME);
+	} while (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
+
+	/* continue handshake, fall through */
+end:
+	free(name);
+	return ret;
+}
+
+#define MAX_ALPN_PROTOCOLS 16
 gnutls_session_t initialize_session(int dtls)
 {
 	gnutls_session_t session;
 	int ret;
+	unsigned i;
 	const char *err;
-
-	if (priorities == NULL)
-		priorities = "NORMAL";
+	gnutls_datum_t alpn[MAX_ALPN_PROTOCOLS];
+	unsigned alpn_size;
+	unsigned flags = GNUTLS_SERVER | GNUTLS_POST_HANDSHAKE_AUTH;
 
 	if (dtls)
-		gnutls_init(&session, GNUTLS_SERVER | GNUTLS_DATAGRAM);
-	else
-		gnutls_init(&session, GNUTLS_SERVER);
+		flags |= GNUTLS_DATAGRAM;
+
+	if (earlydata)
+		flags |= GNUTLS_ENABLE_EARLY_DATA;
+
+	gnutls_init(&session, flags);
 
 	/* allow the use of private ciphersuites.
 	 */
 	gnutls_handshake_set_private_extensions(session, 1);
 
+	gnutls_handshake_set_timeout(session,
+				     GNUTLS_DEFAULT_HANDSHAKE_TIMEOUT);
+
 	if (nodb == 0) {
 		gnutls_db_set_retrieve_function(session, wrap_db_fetch);
 		gnutls_db_set_remove_function(session, wrap_db_delete);
@@ -337,14 +404,48 @@ gnutls_session_t initialize_session(int dtls)
 		gnutls_db_set_ptr(session, NULL);
 	}
 
-#ifdef ENABLE_SESSION_TICKETS
 	if (noticket == 0)
 		gnutls_session_ticket_enable_server(session,
 						    &session_ticket_key);
-#endif
 
-	if (gnutls_priority_set_direct(session, priorities, &err) < 0) {
-		fprintf(stderr, "Syntax error at: %s\n", err);
+	if (earlydata) {
+		gnutls_anti_replay_enable(session, anti_replay);
+		if (HAVE_OPT(MAXEARLYDATA)) {
+			ret = gnutls_record_set_max_early_data_size(session, OPT_VALUE_MAXEARLYDATA);
+			if (ret < 0) {
+				fprintf(stderr, "Could not set max early data size: %s\n", gnutls_strerror(ret));
+				exit(1);
+			}
+		}
+	}
+
+	if (sni_hostname != NULL)
+		gnutls_handshake_set_post_client_hello_function(session,
+								&post_client_hello);
+
+	if (priorities == NULL) {
+		ret = gnutls_set_default_priority(session);
+		if (ret < 0) {
+			fprintf(stderr, "Could not set default policy: %s\n", gnutls_strerror(ret));
+			exit(1);
+		}
+	} else {
+		ret = gnutls_priority_set_direct(session, priorities, &err);
+		if (ret < 0) {
+			fprintf(stderr, "Syntax error at: %s\n", err);
+			exit(1);
+		}
+	}
+
+	alpn_size = MIN(MAX_ALPN_PROTOCOLS,alpn_protos_size);
+	for (i=0;i<alpn_size;i++) {
+		alpn[i].data = (void*)alpn_protos[i];
+		alpn[i].size = strlen(alpn_protos[i]);
+	}
+
+	ret = gnutls_alpn_set_protocols(session, alpn, alpn_size, HAVE_OPT(ALPN_FATAL)?GNUTLS_ALPN_MANDATORY:0);
+	if (ret < 0) {
+		fprintf(stderr, "Error setting ALPN protocols: %s\n", gnutls_strerror(ret));
 		exit(1);
 	}
 
@@ -381,21 +482,21 @@ gnutls_session_t initialize_session(int dtls)
 					GNUTLS_HB_PEER_ALLOWED_TO_SEND);
 
 #ifdef ENABLE_DTLS_SRTP
-        if (HAVE_OPT(SRTP_PROFILES)) {
-                ret =
-                    gnutls_srtp_set_profile_direct(session,
-                                                   OPT_ARG(SRTP_PROFILES),
-                                                   &err);
-                if (ret == GNUTLS_E_INVALID_REQUEST)
-                        fprintf(stderr, "Syntax error at: %s\n", err);
-                else if (ret != 0)
-                        fprintf(stderr, "Error in profiles: %s\n",
-                                gnutls_strerror(ret));
-                else fprintf(stderr,"DTLS profile set to %s\n",
-                             OPT_ARG(SRTP_PROFILES));
+	if (HAVE_OPT(SRTP_PROFILES)) {
+		ret =
+		    gnutls_srtp_set_profile_direct(session,
+						   OPT_ARG(SRTP_PROFILES),
+						   &err);
+		if (ret == GNUTLS_E_INVALID_REQUEST)
+			fprintf(stderr, "Syntax error at: %s\n", err);
+		else if (ret != 0)
+			fprintf(stderr, "Error in profiles: %s\n",
+				gnutls_strerror(ret));
+		else fprintf(stderr,"DTLS profile set to %s\n",
+			     OPT_ARG(SRTP_PROFILES));
 
-                if (ret != 0) exit(1);
-        }
+		if (ret != 0) exit(1);
+	}
 #endif
 
 
@@ -419,10 +520,11 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 	const char *tmp;
 	unsigned char sesid[32];
 	size_t i, sesid_size;
-	char *http_buffer;
+	char *http_buffer, *desc;
 	gnutls_kx_algorithm_t kx_alg;
 	size_t len = 20 * 1024 + strlen(header);
 	char *crtinfo = NULL, *crtinfo_old = NULL;
+	gnutls_protocol_t version;
 	size_t ncrtinfo = 0;
 
 	if (verbose == 0) {
@@ -449,7 +551,7 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 		    gnutls_certificate_get_peers(session, &cert_list_size);
 
 		for (i = 0; i < cert_list_size; i++) {
-			gnutls_x509_crt_t cert;
+			gnutls_x509_crt_t cert = NULL;
 			gnutls_datum_t info;
 
 			if (gnutls_x509_crt_init(&cert) == 0 &&
@@ -479,6 +581,7 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 				crtinfo[ncrtinfo] = '\0';
 				gnutls_free(info.data);
 			}
+			gnutls_x509_crt_deinit(cert);
 		}
 	}
 
@@ -490,6 +593,8 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 
 	strcpy(http_buffer, HTTP_BEGIN);
 
+	version = gnutls_protocol_get_version(session);
+
 	/* print session_id */
 	sesid_size = sizeof(sesid);
 	gnutls_session_get_id(session, sesid, &sesid_size);
@@ -498,7 +603,7 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 		snprintf(tmp_buffer, tmp_buffer_size, "%.2X", sesid[i]);
 	snprintf(tmp_buffer, tmp_buffer_size, "</i></p>\n");
 	snprintf(tmp_buffer, tmp_buffer_size,
-		 "<h5>If your browser supports session resuming, then you should see the "
+		 "<h5>If your browser supports session resumption, then you should see the "
 		 "same session ID, when you press the <b>reload</b> button.</h5>\n");
 
 	/* Here unlike print_info() we use the kx algorithm to distinguish
@@ -529,39 +634,35 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 #endif
 
 #ifdef ENABLE_PSK
-	if (kx_alg == GNUTLS_KX_PSK) {
+	if (kx_alg == GNUTLS_KX_PSK && gnutls_psk_server_get_username(session)) {
 		snprintf(tmp_buffer, tmp_buffer_size,
 			 "<p>Connected as user '%s'.</p>\n",
 			 gnutls_psk_server_get_username(session));
 	}
 #endif
 
-#ifdef ENABLE_ANON
-	if (kx_alg == GNUTLS_KX_ANON_DH) {
-		snprintf(tmp_buffer, tmp_buffer_size,
-			 "<p> Connect using anonymous DH (prime of %d bits)</p>\n",
-			 gnutls_dh_get_prime_bits(session));
-	}
-#endif
-
-	if (kx_alg == GNUTLS_KX_DHE_RSA || kx_alg == GNUTLS_KX_DHE_DSS) {
-		snprintf(tmp_buffer, tmp_buffer_size,
-			 "Ephemeral DH using prime of <b>%d</b> bits.<br>\n",
-			 gnutls_dh_get_prime_bits(session));
-	}
 
 	/* print session information */
 	strcat(http_buffer, "<P>\n");
 
 	tmp =
-	    gnutls_protocol_get_name(gnutls_protocol_get_version(session));
+	    gnutls_protocol_get_name(version);
 	if (tmp == NULL)
 		tmp = str_unknown;
 	snprintf(tmp_buffer, tmp_buffer_size,
 		 "<TABLE border=1><TR><TD>Protocol version:</TD><TD>%s</TD></TR>\n",
 		 tmp);
 
-	if (gnutls_auth_get_type(session) == GNUTLS_CRD_CERTIFICATE) {
+	desc = gnutls_session_get_desc(session);
+	if (desc) {
+		snprintf(tmp_buffer, tmp_buffer_size,
+			"<TR><TD>Description:</TD><TD>%s</TD></TR>\n",
+			desc);
+		gnutls_free(desc);
+	}
+
+	if (gnutls_auth_get_type(session) == GNUTLS_CRD_CERTIFICATE &&
+	    gnutls_certificate_type_get(session) != GNUTLS_CRT_X509) {
 		tmp =
 		    gnutls_certificate_type_get_name
 		    (gnutls_certificate_type_get(session));
@@ -572,17 +673,42 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 			 tmp);
 	}
 
-	tmp = gnutls_kx_get_name(kx_alg);
-	if (tmp == NULL)
-		tmp = str_unknown;
-	snprintf(tmp_buffer, tmp_buffer_size,
-		 "<TR><TD>Key Exchange:</TD><TD>%s</TD></TR>\n", tmp);
+	if (version < GNUTLS_TLS1_3) {
+		tmp = gnutls_kx_get_name(kx_alg);
+		if (tmp == NULL)
+			tmp = str_unknown;
+		snprintf(tmp_buffer, tmp_buffer_size,
+			 "<TR><TD>Key Exchange:</TD><TD>%s</TD></TR>\n", tmp);
 
-	tmp = gnutls_compression_get_name(gnutls_compression_get(session));
-	if (tmp == NULL)
-		tmp = str_unknown;
-	snprintf(tmp_buffer, tmp_buffer_size,
-		 "<TR><TD>Compression</TD><TD>%s</TD></TR>\n", tmp);
+#ifdef ENABLE_ANON
+		if (kx_alg == GNUTLS_KX_ANON_DH) {
+			snprintf(tmp_buffer, tmp_buffer_size,
+				 "<p> Connect using anonymous DH (prime of %d bits)</p>\n",
+				 gnutls_dh_get_prime_bits(session));
+		}
+#endif
+
+		if (kx_alg == GNUTLS_KX_DHE_RSA || kx_alg == GNUTLS_KX_DHE_DSS) {
+			snprintf(tmp_buffer, tmp_buffer_size,
+				 "Ephemeral DH using prime of <b>%d</b> bits.<br>\n",
+				 gnutls_dh_get_prime_bits(session));
+		}
+
+		tmp = gnutls_compression_get_name(gnutls_compression_get(session));
+		if (tmp == NULL)
+			tmp = str_unknown;
+		snprintf(tmp_buffer, tmp_buffer_size,
+			 "<TR><TD>Compression</TD><TD>%s</TD></TR>\n", tmp);
+
+		tmp = gnutls_cipher_suite_get_name(kx_alg,
+						   gnutls_cipher_get(session),
+						   gnutls_mac_get(session));
+		if (tmp == NULL)
+			tmp = str_unknown;
+		snprintf(tmp_buffer, tmp_buffer_size,
+			 "<TR><TD>Ciphersuite</TD><TD>%s</TD></TR></p></TABLE>\n",
+			 tmp);
+	}
 
 	tmp = gnutls_cipher_get_name(gnutls_cipher_get(session));
 	if (tmp == NULL)
@@ -596,14 +722,6 @@ static char *peer_print_info(gnutls_session_t session, int *ret_length,
 	snprintf(tmp_buffer, tmp_buffer_size,
 		 "<TR><TD>MAC</TD><TD>%s</TD></TR>\n", tmp);
 
-	tmp = gnutls_cipher_suite_get_name(kx_alg,
-					   gnutls_cipher_get(session),
-					   gnutls_mac_get(session));
-	if (tmp == NULL)
-		tmp = str_unknown;
-	snprintf(tmp_buffer, tmp_buffer_size,
-		 "<TR><TD>Ciphersuite</TD><TD>%s</TD></TR></p></TABLE>\n",
-		 tmp);
 
 	if (crtinfo) {
 		snprintf(tmp_buffer, tmp_buffer_size,
@@ -627,7 +745,7 @@ const char *human_addr(const struct sockaddr *sa, socklen_t salen,
 	size_t l;
 
 	if (!buf || !buflen)
-		return NULL;
+		return "(error)";
 
 	*buf = 0;
 
@@ -649,8 +767,8 @@ const char *human_addr(const struct sockaddr *sa, socklen_t salen,
 
 	if (getnameinfo(sa, salen, buf, buflen, NULL, 0, NI_NUMERICHOST) !=
 	    0) {	
-		   return NULL;
-        }
+		return "(error)";
+	}
 
 	l = strlen(buf);
 	buf += l;
@@ -666,7 +784,7 @@ const char *human_addr(const struct sockaddr *sa, socklen_t salen,
 	if (getnameinfo(sa, salen, NULL, 0, buf, buflen, NI_NUMERICSERV) !=
 	    0) {
 		snprintf(buf, buflen, "%s", " unknown");
-        }
+	}
 
 	return save_buf;
 }
@@ -761,7 +879,7 @@ int listen_socket(const char *name, int listen_port, int socktype)
 			yes = 1;
 			/* avoid listen on ipv6 addresses failing
 			 * because already listening on ipv4 addresses: */
-			setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY,
+			(void)setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY,
 				   (const void *) &yes, sizeof(yes));
 		}
 #endif
@@ -830,7 +948,7 @@ static void strip(char *data)
 
 	for (i = 0; i < len; i++) {
 		if (data[i] == '\r' && data[i + 1] == '\n'
-		    && data[i + 1] == 0) {
+		    && data[i + 2] == 0) {
 			data[i] = '\n';
 			data[i + 1] = 0;
 			break;
@@ -863,15 +981,27 @@ get_response(gnutls_session_t session, char *request,
 	if (http != 0) {
 		*response = peer_print_info(session, response_length, h);
 	} else {
+		int ret;
 		strip(request);
-		fprintf(stderr, "received: %s\n", request);
-		if (check_command(session, request)) {
-			*response = NULL;
-			*response_length = 0;
+		fprintf(stderr, "received cmd: %s\n", request);
+
+		ret = check_command(session, request, disable_client_cert);
+		if (ret > 0) {
+			*response = strdup("Successfully executed command\n");
+			if (*response == NULL) {
+				fprintf(stderr, "Memory error\n");
+				exit(1);
+			}
+			*response_length = strlen(*response);
 			return;
+		} else if (ret == 0) {
+			*response = strdup(request);
+			*response_length = ((*response) ? strlen(*response) : 0);
+		} else {
+			do {
+				ret = gnutls_alert_send(session, GNUTLS_AL_FATAL, GNUTLS_A_UNEXPECTED_MESSAGE);
+			} while(ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
 		}
-		*response = strdup(request);
-		*response_length = ((*response) ? strlen(*response) : 0);
 	}
 
 	return;
@@ -920,6 +1050,8 @@ int main(int argc, char **argv)
 	int ret, mtu, port;
 	char name[256];
 	int cert_set = 0;
+	unsigned use_static_dh_params = 0;
+	unsigned i;
 
 	cmd_parser(argc, argv);
 
@@ -983,7 +1115,7 @@ int main(int argc, char **argv)
 	} else if (dh_params_file) {
 		read_dh_params();
 	} else {
-		static_dh_params();
+		use_static_dh_params = 1;
 	}
 
 	if (gnutls_certificate_allocate_credentials(&cert_cred) < 0) {
@@ -1013,101 +1145,52 @@ int main(int argc, char **argv)
 			printf("Processed %d CRL(s).\n", ret);
 		}
 	}
-#ifdef ENABLE_OPENPGP
-	if (pgp_keyring != NULL) {
-		ret =
-		    gnutls_certificate_set_openpgp_keyring_file(cert_cred,
-								pgp_keyring,
-								GNUTLS_OPENPGP_FMT_BASE64);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error setting the OpenPGP keyring file\n");
-			GERR(ret);
+
+	if (x509_certfile_size > 0 && x509_keyfile_size > 0) {
+		for (i = 0; i < x509_certfile_size; i++) {
+			ret = gnutls_certificate_set_x509_key_file
+			    (cert_cred, x509_certfile[i], x509_keyfile[i], x509ctype);
+			if (ret < 0) {
+				fprintf(stderr,
+						"Error reading '%s' or '%s'\n",
+						x509_certfile[i], x509_keyfile[i]);
+				GERR(ret);
+				exit(1);
+			} else
+				cert_set = 1;
 		}
 	}
 
-	if (pgp_certfile != NULL && pgp_keyfile != NULL) {
-		if (HAVE_OPT(PGPSUBKEY))
-			ret = gnutls_certificate_set_openpgp_key_file2
-			    (cert_cred, pgp_certfile, pgp_keyfile,
-			     OPT_ARG(PGPSUBKEY),
-			     GNUTLS_OPENPGP_FMT_BASE64);
-		else
-			ret = gnutls_certificate_set_openpgp_key_file
-			    (cert_cred, pgp_certfile, pgp_keyfile,
-			     GNUTLS_OPENPGP_FMT_BASE64);
-
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error[%d] while reading the OpenPGP key pair ('%s', '%s')\n",
-				ret, pgp_certfile, pgp_keyfile);
-			GERR(ret);
-		} else
-			cert_set = 1;
-	}
-#endif
-
-	if (x509_certfile != NULL && x509_keyfile != NULL) {
-		ret = gnutls_certificate_set_x509_key_file
-		    (cert_cred, x509_certfile, x509_keyfile, x509ctype);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error reading '%s' or '%s'\n",
-				x509_certfile, x509_keyfile);
-			GERR(ret);
-			exit(1);
-		} else
-			cert_set = 1;
-	}
-
-	if (x509_dsacertfile != NULL && x509_dsakeyfile != NULL) {
-		ret = gnutls_certificate_set_x509_key_file
-		    (cert_cred, x509_dsacertfile, x509_dsakeyfile,
-		     x509ctype);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error reading '%s' or '%s'\n",
-				x509_dsacertfile, x509_dsakeyfile);
-			GERR(ret);
-			exit(1);
-		} else
-			cert_set = 1;
-	}
-
-	if (x509_ecccertfile != NULL && x509_ecckeyfile != NULL) {
-		ret = gnutls_certificate_set_x509_key_file
-		    (cert_cred, x509_ecccertfile, x509_ecckeyfile,
-		     x509ctype);
-		if (ret < 0) {
-			fprintf(stderr,
-				"Error reading '%s' or '%s'\n",
-				x509_ecccertfile, x509_ecckeyfile);
-			GERR(ret);
-			exit(1);
-		} else
-			cert_set = 1;
-	}
-
 	if (cert_set == 0) {
 		fprintf(stderr,
 			"Warning: no private key and certificate pairs were set.\n");
 	}
 
 	/* OCSP status-request TLS extension */
-	if (status_response_ocsp) {
-		if (gnutls_certificate_set_ocsp_status_request_file
-		    (cert_cred, status_response_ocsp, 0) < 0) {
+	if (HAVE_OPT(IGNORE_OCSP_RESPONSE_ERRORS))
+		gnutls_certificate_set_flags(cert_cred, GNUTLS_CERTIFICATE_SKIP_OCSP_RESPONSE_CHECK);
+
+	for (i = 0; i < ocsp_responses_size; i++ ) {
+		ret = gnutls_certificate_set_ocsp_status_request_file
+		    (cert_cred, ocsp_responses[i], 0);
+		if (ret < 0) {
 			fprintf(stderr,
-				"Cannot set OCSP status request file: %s\n",
+				"Cannot set OCSP status request file: %s: %s\n",
+				ocsp_responses[i],
 				gnutls_strerror(ret));
 			exit(1);
 		}
 	}
 
-	gnutls_certificate_set_params_function(cert_cred, get_params);
-/*     gnutls_certificate_set_dh_params(cert_cred, dh_params);
- *     gnutls_certificate_set_rsa_export_params(cert_cred, rsa_params);
- */
+	if (use_static_dh_params) {
+		ret = gnutls_certificate_set_known_dh_params(cert_cred, GNUTLS_SEC_PARAM_MEDIUM);
+		if (ret < 0) {
+			fprintf(stderr, "Error while setting DH parameters: %s\n", gnutls_strerror(ret));
+			exit(1);
+		}
+	} else {
+		gnutls_certificate_set_params_function(cert_cred, get_params);
+	}
 
 	/* this is a password file (created with the included srpcrypt utility) 
 	 * Read README.crypt prior to using SRP.
@@ -1158,22 +1241,45 @@ int main(int argc, char **argv)
 			}
 		}
 
-		gnutls_psk_set_server_params_function(psk_cred,
-						      get_params);
+		if (use_static_dh_params) {
+			ret = gnutls_psk_set_server_known_dh_params(psk_cred, GNUTLS_SEC_PARAM_MEDIUM);
+			if (ret < 0) {
+				fprintf(stderr, "Error while setting DH parameters: %s\n", gnutls_strerror(ret));
+				exit(1);
+			}
+		} else {
+			gnutls_psk_set_server_params_function(psk_cred,
+							      get_params);
+		}
 	}
 #endif
 
 #ifdef ENABLE_ANON
 	gnutls_anon_allocate_server_credentials(&dh_cred);
-	gnutls_anon_set_server_params_function(dh_cred, get_params);
 
-/*      gnutls_anon_set_server_dh_params(dh_cred, dh_params); */
+	if (use_static_dh_params) {
+		ret = gnutls_anon_set_server_known_dh_params(dh_cred, GNUTLS_SEC_PARAM_MEDIUM);
+		if (ret < 0) {
+			fprintf(stderr, "Error while setting DH parameters: %s\n", gnutls_strerror(ret));
+			exit(1);
+		}
+	} else {
+		gnutls_anon_set_server_params_function(dh_cred, get_params);
+	}
 #endif
 
-#ifdef ENABLE_SESSION_TICKETS
 	if (noticket == 0)
 		gnutls_session_ticket_key_generate(&session_ticket_key);
-#endif
+
+	if (earlydata) {
+		ret = gnutls_anti_replay_init(&anti_replay);
+		if (ret < 0) {
+			fprintf(stderr, "Error while initializing anti-replay: %s\n", gnutls_strerror(ret));
+			exit(1);
+		}
+		gnutls_anti_replay_set_add_function(anti_replay, anti_replay_db_add);
+		gnutls_anti_replay_set_ptr(anti_replay, NULL);
+	}
 
 	if (HAVE_OPT(MTU))
 		mtu = OPT_VALUE_MTU;
@@ -1204,12 +1310,12 @@ static void retry_handshake(listener_item *j)
 	} else if (r < 0) {
 		j->http_state = HTTP_STATE_CLOSING;
 		check_alert(j->tls_session, r);
-		fprintf(stderr, "Error in handshake\n");
-		GERR(r);
+		fprintf(stderr, "Error in handshake: %s\n", gnutls_strerror(r));
 
 		do {
 			ret = gnutls_alert_send_appropriate(j->tls_session, r);
 		} while (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
+		j->close_ok = 0;
 	} else if (r == 0) {
 		if (gnutls_session_is_resumed(j->tls_session) != 0 && verbose != 0)
 			printf("*** This is a resumed session\n");
@@ -1227,14 +1333,16 @@ static void retry_handshake(listener_item *j)
 			print_info(j->tls_session, verbose, verbose);
 		}
 
+		j->close_ok = 1;
 		j->handshake_ok = 1;
 	}
 }
 
 static void try_rehandshake(listener_item *j)
 {
-int r, ret;
+	int r, ret;
 	fprintf(stderr, "*** Received hello message\n");
+
 	do {
 		r = gnutls_handshake(j->tls_session);
 	} while (r == GNUTLS_E_INTERRUPTED || r == GNUTLS_E_AGAIN);
@@ -1243,9 +1351,10 @@ int r, ret;
 		do {
 			ret = gnutls_alert_send_appropriate(j->tls_session, r);
 		} while (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
-		GERR(r);
+		fprintf(stderr, "Error in rehandshake: %s\n", gnutls_strerror(r));
 		j->http_state = HTTP_STATE_CLOSING;
 	} else {
+		j->close_ok = 1;
 		j->http_state = HTTP_STATE_REQUEST;
 	}
 }
@@ -1353,6 +1462,7 @@ static void tcp_server(const char *name, int port)
 					    (j->tls_session, accept_fd);
 					set_read_funcs(j->tls_session);
 					j->handshake_ok = 0;
+					j->close_ok = 0;
 
 					if (verbose != 0) {
 						ctt = ctime(&tt);
@@ -1375,7 +1485,7 @@ static void tcp_server(const char *name, int port)
 
 			if (FD_ISSET(j->fd, &rd) && !j->listen_socket) {
 /* read partial GET request */
-				char buf[1024];
+				char buf[16*1024];
 				int r;
 
 				if (j->handshake_ok == 0) {
@@ -1383,25 +1493,46 @@ static void tcp_server(const char *name, int port)
 				}
 
 				if (j->handshake_ok == 1) {
-					r = gnutls_record_recv(j->
-							       tls_session,
-							       buf,
-							       MIN(1024,
-								   SMALL_READ_TEST));
+					int earlydata_read = 0;
+					if (earlydata && !j->earlydata_eof) {
+						r = gnutls_record_recv_early_data(j->
+										  tls_session,
+										  buf,
+										  MIN(sizeof(buf),
+										      SMALL_READ_TEST));
+						if (r == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
+							j->earlydata_eof = 1;
+						}
+						if (r == 0) {
+							earlydata_read = 1;
+						}
+					}
+					if (!earlydata_read) {
+						r = gnutls_record_recv(j->
+								       tls_session,
+								       buf,
+								       MIN(sizeof(buf),
+									   SMALL_READ_TEST));
+					}
 					if (r == GNUTLS_E_INTERRUPTED || r == GNUTLS_E_AGAIN) {
 						/* do nothing */
 					} else if (r <= 0) {
 						if (r == GNUTLS_E_HEARTBEAT_PING_RECEIVED) {
 							gnutls_heartbeat_pong(j->tls_session, 0);
 						} else if (r == GNUTLS_E_REHANDSHAKE) {
-						    	try_rehandshake(j);
+							try_rehandshake(j);
 						} else {
 							j->http_state = HTTP_STATE_CLOSING;
 							if (r < 0) {
+								int ret;
 								check_alert(j->tls_session, r);
 								fprintf(stderr,
 								     "Error while receiving data\n");
+								do {
+									ret = gnutls_alert_send_appropriate(j->tls_session, r);
+								} while (ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED);
 								GERR(r);
+								j->close_ok = 0;
 							}
 						}
 					} else {
@@ -1501,6 +1632,7 @@ static void tcp_server(const char *name, int port)
 								j->http_state = HTTP_STATE_REQUEST;
 								free(j->
 								     http_response);
+								j->http_response = NULL;
 								j->response_length = 0;
 								j->request_length = 0;
 								j->http_request[0] = 0;
@@ -1545,6 +1677,9 @@ static void tcp_server(const char *name, int port)
 	if (noticket == 0)
 		gnutls_free(session_ticket_key.data);
 
+	if (earlydata)
+		gnutls_anti_replay_deinit(anti_replay);
+
 	if (nodb == 0)
 		wrap_db_deinit();
 	gnutls_global_deinit();
@@ -1573,6 +1708,7 @@ static void cmd_parser(int argc, char **argv)
 
 	nodb = HAVE_OPT(NODB);
 	noticket = HAVE_OPT(NOTICKET);
+	earlydata = HAVE_OPT(EARLYDATA);
 
 	if (HAVE_OPT(ECHO))
 		http = 0;
@@ -1589,35 +1725,31 @@ static void cmd_parser(int argc, char **argv)
 	if (HAVE_OPT(DHPARAMS))
 		dh_params_file = OPT_ARG(DHPARAMS);
 
-	if (HAVE_OPT(X509KEYFILE))
-		x509_keyfile = OPT_ARG(X509KEYFILE);
-	if (HAVE_OPT(X509CERTFILE))
-		x509_certfile = OPT_ARG(X509CERTFILE);
+	if (HAVE_OPT(ALPN)) {
+		alpn_protos = STACKLST_OPT(ALPN);
+		alpn_protos_size = STACKCT_OPT(ALPN);
+	}
 
-	if (HAVE_OPT(X509DSAKEYFILE))
-		x509_dsakeyfile = OPT_ARG(X509DSAKEYFILE);
-	if (HAVE_OPT(X509DSACERTFILE))
-		x509_dsacertfile = OPT_ARG(X509DSACERTFILE);
+	if (HAVE_OPT(X509KEYFILE)) {
+		x509_keyfile = STACKLST_OPT(X509KEYFILE);
+		x509_keyfile_size = STACKCT_OPT(X509KEYFILE);
+	}
 
+	if (HAVE_OPT(X509CERTFILE)) {
+		x509_certfile = STACKLST_OPT(X509CERTFILE);
+		x509_certfile_size = STACKCT_OPT(X509CERTFILE);
+	}
 
-	if (HAVE_OPT(X509ECCKEYFILE))
-		x509_ecckeyfile = OPT_ARG(X509ECCKEYFILE);
-	if (HAVE_OPT(X509ECCCERTFILE))
-		x509_ecccertfile = OPT_ARG(X509ECCCERTFILE);
+	if (x509_certfile_size != x509_keyfile_size) {
+		fprintf(stderr, "The certificate number provided (%u) doesn't match the keys (%u)\n",
+			x509_certfile_size, x509_keyfile_size);
+	}
 
 	if (HAVE_OPT(X509CAFILE))
 		x509_cafile = OPT_ARG(X509CAFILE);
 	if (HAVE_OPT(X509CRLFILE))
 		x509_crlfile = OPT_ARG(X509CRLFILE);
 
-	if (HAVE_OPT(PGPKEYFILE))
-		pgp_keyfile = OPT_ARG(PGPKEYFILE);
-	if (HAVE_OPT(PGPCERTFILE))
-		pgp_certfile = OPT_ARG(PGPCERTFILE);
-
-	if (HAVE_OPT(PGPKEYRING))
-		pgp_keyring = OPT_ARG(PGPKEYRING);
-
 	if (HAVE_OPT(SRPPASSWD))
 		srp_passwd = OPT_ARG(SRPPASSWD);
 	if (HAVE_OPT(SRPPASSWDCONF))
@@ -1626,57 +1758,101 @@ static void cmd_parser(int argc, char **argv)
 	if (HAVE_OPT(PSKPASSWD))
 		psk_passwd = OPT_ARG(PSKPASSWD);
 
-	if (HAVE_OPT(OCSP_RESPONSE))
-		status_response_ocsp = OPT_ARG(OCSP_RESPONSE);
+	if (HAVE_OPT(OCSP_RESPONSE)) {
+		ocsp_responses = STACKLST_OPT(OCSP_RESPONSE);
+		ocsp_responses_size = STACKCT_OPT(OCSP_RESPONSE);
+	}
+
+	if (HAVE_OPT(SNI_HOSTNAME))
+		sni_hostname = OPT_ARG(SNI_HOSTNAME);
+
+	if (HAVE_OPT(SNI_HOSTNAME_FATAL))
+		sni_hostname_fatal = 1;
 
 }
 
 /* session resuming support */
 
-#define SESSION_ID_SIZE 32
-#define SESSION_DATA_SIZE 1024
+#define SESSION_ID_SIZE 128
+#define SESSION_DATA_SIZE (16*1024)
 
 typedef struct {
-	char session_id[SESSION_ID_SIZE];
+	unsigned char session_id[SESSION_ID_SIZE];
 	unsigned int session_id_size;
 
-	char session_data[SESSION_DATA_SIZE];
-	unsigned int session_data_size;
+	gnutls_datum_t session_data;
 } CACHE;
 
 static CACHE *cache_db;
-int cache_db_ptr = 0;
+static int cache_db_ptr;
+static int cache_db_alloc;
 
 static void wrap_db_init(void)
 {
-	/* allocate cache_db */
-	cache_db = calloc(1, ssl_session_cache * sizeof(CACHE));
 }
 
 static void wrap_db_deinit(void)
 {
+	int i;
+
+	for (i = 0; i < cache_db_ptr; i++)
+		free(cache_db[i].session_data.data);
+	free(cache_db);
 }
 
 static int
 wrap_db_store(void *dbf, gnutls_datum_t key, gnutls_datum_t data)
 {
-
-	if (cache_db == NULL)
-		return -1;
+	int i;
+	time_t now = time(0);
 
 	if (key.size > SESSION_ID_SIZE)
-		return -1;
+		return GNUTLS_E_DB_ERROR;
 	if (data.size > SESSION_DATA_SIZE)
-		return -1;
+		return GNUTLS_E_DB_ERROR;
 
-	memcpy(cache_db[cache_db_ptr].session_id, key.data, key.size);
-	cache_db[cache_db_ptr].session_id_size = key.size;
+	if (cache_db_ptr < cache_db_alloc)
+		i = cache_db_ptr++;
+	else {
+		/* find empty or expired slot to store the new entry */
+		for (i = 0; i < cache_db_ptr; i++)
+			if (cache_db[i].session_id_size == 0 ||
+			    !(now <
+			      gnutls_db_check_entry_expire_time(&cache_db[i].
+								session_data)))
+				break;
 
-	memcpy(cache_db[cache_db_ptr].session_data, data.data, data.size);
-	cache_db[cache_db_ptr].session_data_size = data.size;
+		if (i == cache_db_ptr) {
+			/* try to allocate additional slots */
+			if (cache_db_ptr == ssl_session_cache) {
+				fprintf(stderr,
+					"Error: too many sessions\n");
+				return GNUTLS_E_DB_ERROR;
+			}
+			cache_db_alloc = cache_db_alloc * 2 + 1;
+			cache_db = realloc(cache_db,
+					   cache_db_alloc * sizeof(CACHE));
+			if (!cache_db)
+				return GNUTLS_E_MEMORY_ERROR;
+			memset(cache_db + cache_db_ptr, 0,
+			       (cache_db_alloc - cache_db_ptr) * sizeof(CACHE));
+			cache_db_ptr++;
+		}
+	}
 
-	cache_db_ptr++;
-	cache_db_ptr %= ssl_session_cache;
+	memcpy(cache_db[i].session_id, key.data, key.size);
+	cache_db[i].session_id_size = key.size;
+
+	/* resize the data slot if needed */
+	if (cache_db[i].session_data.size < data.size) {
+		cache_db[i].session_data.data =
+			realloc(cache_db[i].session_data.data,
+				data.size);
+		if (!cache_db[i].session_data.data)
+			return GNUTLS_E_MEMORY_ERROR;
+	}
+	memcpy(cache_db[i].session_data.data, data.data, data.size);
+	cache_db[i].session_data.size = data.size;
 
 	return 0;
 }
@@ -1684,22 +1860,22 @@ wrap_db_store(void *dbf, gnutls_datum_t key, gnutls_datum_t data)
 static gnutls_datum_t wrap_db_fetch(void *dbf, gnutls_datum_t key)
 {
 	gnutls_datum_t res = { NULL, 0 };
+	time_t now = time(0);
 	int i;
 
-	if (cache_db == NULL)
-		return res;
-
-	for (i = 0; i < ssl_session_cache; i++) {
+	for (i = 0; i < cache_db_ptr; i++) {
 		if (key.size == cache_db[i].session_id_size &&
 		    memcmp(key.data, cache_db[i].session_id,
-			   key.size) == 0) {
-			res.size = cache_db[i].session_data_size;
+			   key.size) == 0 &&
+		    now < gnutls_db_check_entry_expire_time(&cache_db[i].
+							    session_data)) {
+			res.size = cache_db[i].session_data.size;
 
-			res.data = gnutls_malloc(res.size);
+			res.data = malloc(res.size);
 			if (res.data == NULL)
 				return res;
 
-			memcpy(res.data, cache_db[i].session_data,
+			memcpy(res.data, cache_db[i].session_data.data,
 			       res.size);
 
 			return res;
@@ -1712,20 +1888,37 @@ static int wrap_db_delete(void *dbf, gnutls_datum_t key)
 {
 	int i;
 
-	if (cache_db == NULL)
-		return -1;
-
-	for (i = 0; i < ssl_session_cache; i++) {
-		if (key.size == (unsigned int) cache_db[i].session_id_size
-		    && memcmp(key.data, cache_db[i].session_id,
-			      key.size) == 0) {
+	for (i = 0; i < cache_db_ptr; i++) {
+		if (key.size == cache_db[i].session_id_size &&
+		    memcmp(key.data, cache_db[i].session_id,
+			   key.size) == 0) {
 
 			cache_db[i].session_id_size = 0;
-			cache_db[i].session_data_size = 0;
+			free(cache_db[i].session_data.data);
+			cache_db[i].session_data.data = NULL;
+			cache_db[i].session_data.size = 0;
 
 			return 0;
 		}
 	}
 
-	return -1;
+	return GNUTLS_E_DB_ERROR;
+}
+
+static int
+anti_replay_db_add(void *dbf, time_t exp, const gnutls_datum_t *key, const gnutls_datum_t *data)
+{
+	time_t now = time(0);
+	int i;
+
+	for (i = 0; i < cache_db_ptr; i++) {
+		if (key->size == cache_db[i].session_id_size &&
+		    memcmp(key->data, cache_db[i].session_id,
+			   key->size) == 0 &&
+		    now < gnutls_db_check_entry_expire_time(&cache_db[i].
+							    session_data))
+			return GNUTLS_E_DB_ENTRY_EXISTS;
+	}
+
+	return wrap_db_store(dbf, *key, *data);
 }
